// Copyright (c) 2023 Citrix Systems, Inc. All Rights Reserved.
// You may only reproduce, distribute, perform, display, or prepare derivative works of this file pursuant to a valid license from Citrix.
// THIS SAMPLE CODE IS PROVIDED BY CITRIX "AS IS" AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.

// CredInjectSample.cpp
//
// A command line utility for testing credential insertion.
//
// This sample demonstrates how to use the Credential Insertion SDK.
// The code processes the command line, calls LogonSsoUser(),
// LogonSsoUserWithPin() or LogoffSsoUser() as appropriate and handles errors.
//
// Note that this sample only supports the insertion/removal of credentials
// (either PIN or user-domain-password) to/from the SSOn cache.
// For this example to work, AM must be running.
// The logoff (-L) option will pop credentials from the SSOn stack, but
// does not remove AM tickets.

#include <iostream>
#include <string>
#include <iconv.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>

#include "CredInject.h"

#define PATH_MAX 4*1024

int debug = 0;

typedef enum { 
    NO_ERROR = 0,
    INVALID_ARGUMENTS = 5,
    IO_ERROR,
} CIError;

typedef wchar_t AMChar;

///--------------------------------------------------------------------------------------------------
/// Purpose:
///     @brief Blanks a password to minimise the risk of 'ps' showing it
///
/// Parameters:
///     @param[in] it - pointer to string to blank
///
/// Returns:
///     @returns void
///--------------------------------------------------------------------------------------------------
void blank_it(char *it)
{
    int i = strlen(it);

    while (i-- > 1)
    {
        it[i] = '\0';
    }

    // Dummy value.
    it[0] = 'X';
}

///--------------------------------------------------------------------------------------------------
/// Purpose:
///     @brief Reads a credential, and blanks the command line
///
/// Returns:
///     @returns string containing the credential
///--------------------------------------------------------------------------------------------------
static std::string readCredential(void)
{
    std::string credential(optarg);
    blank_it(optarg);
    return credential;
}

///--------------------------------------------------------------------------------------------------
/// Purpose:
///     @brief Converts a std::string to AMChar *
///
/// Parameters:
///     @param [in] inBuf - string input buffer
///     @param [out] outBuf - returns a freshly allocated buffer containing a copy of the string,
///                           followed by a zero
/// Returns:
///     @returns none
///--------------------------------------------------------------------------------------------------
void StdStringToAMChar_p(const std::string &inBuf, AMChar **outBuf)
{
    std::wstring wstring;
    wstring.assign(inBuf.begin(), inBuf.end());
    *outBuf = new AMChar[wstring.length() + 1];
    if (*outBuf != NULL)
    {
        wstring.copy(*outBuf, wstring.length());
        (*outBuf)[wstring.length()] = 0;
    }
}

///--------------------------------------------------------------------------------------------------
/// Purpose:
///     @brief Converts an AMChar * buffer to a std::string
///
/// Parameters:
///     @param [in] inBuf - a pointer to the first character of the buffer to be copied
///     @param [out] outString - the new string that has been created
///
/// Returns:
///     @returns none
///--------------------------------------------------------------------------------------------------
void AMChar_pToStdString(const AMChar *inString, std::string& outString)
{
    char dest[PATH_MAX];

    // Use iconv to do the conversion, choose LE encoding which
    // used on x86 and AMChar input is wchar_t which is 32-bits.
    iconv_t cb = iconv_open("UTF-8", "UTF-32LE");
    if (cb == (iconv_t) -1)
    {
        std::cerr << "Unable to perform conversion. iconv_open(\"UTF-8\", \"UTF-32LE\") failed!" << std::endl;
        return ;
    }
    
    size_t inbytesleft = sizeof(wchar_t) *(wcslen(inString)+1);
    char *inbuf = (char *)inString;
    char *outbuf = dest;
    size_t outbytesleft = PATH_MAX;
    size_t length = iconv(cb, &inbuf, &inbytesleft, &outbuf, &outbytesleft);
    
    iconv_close(cb);
    
    if (length == (size_t ) -1)
    {
        std::cerr << "Conversion to string failed: " << strerror(errno) << std::endl;
        return ;
    }
    outString.assign(dest);
}

///--------------------------------------------------------------------------------------------------
/// Purpose:
///     @brief Insert the credentials passed to storebrowse via command line (if any) into AM
///            Username, domain and password must've been passed using the -U, -D and -P options respectively
///            If a PIN is passed, username and domain must be left empty
///
/// Returns:
///     @returns none
///--------------------------------------------------------------------------------------------------
int insertSSOnCredentials(const std::string &userName, const std::string &domain, const std::string &password)
{
    if (debug) std::cout << "insertSSOnCredentials()" << std::endl;

    // Check if we have the credentials from the command line.
    if (password.empty())
    {
        // No credentials available.
        return INVALID_ARGUMENTS;
    }

    // For a PIN only the password is set, otherwise need both the others.
    if (userName.empty() != domain.empty())
    {
        // Inconsistent credentials given
        return INVALID_ARGUMENTS;
    }

    // Credentials are available, inject them to AM.
    wchar_t *password_wide = NULL;
    StdStringToAMChar_p(password, &password_wide);
    CitrixSSOnSDK::LOGONSSOUSER_ERROR_CODE result;
    if (userName.empty())
    {
        result = CitrixSSOnSDK::LogonSsoUserWithPin(password_wide);
    }
    else
    {
        wchar_t *userName_wide = NULL;
        wchar_t *domain_wide   = NULL;
        StdStringToAMChar_p(userName, &userName_wide);
        StdStringToAMChar_p(domain,   &domain_wide);
        result = CitrixSSOnSDK::LogonSsoUser(userName_wide,
                                             domain_wide,
                                             password_wide);
        delete [] domain_wide;
        domain_wide = NULL;
        delete [] userName_wide;
        userName_wide = NULL;
    }
    delete [] password_wide;
    password_wide = NULL;

    // As this is just a simple test utility, do not bother to
    // clear the temporary strings containing the credentials.
    std::string type(userName.empty() ? "PIN" : "Username/Domain/Password credentials");
    if (result != CitrixSSOnSDK::LOGONSSOUSER_OK)
    {
        // Convert the result description to a UTF-8 string.
        std::string result_description;
        AMChar_pToStdString(CitrixSSOnSDK::ErrorDescription(result), result_description);

        std::cerr << "Result from adding " << type << ": " << result_description << std::endl;
    }

    // Check whether the credential injection was successful.
    switch (result)
    {
    case CitrixSSOnSDK::LOGONSSOUSER_OK:
        break;

    case CitrixSSOnSDK::LOGONSSOUSER_INVALID_PARAMETER:
        return INVALID_ARGUMENTS;

    default:
        return IO_ERROR;
    }

    std::cerr << type << " inserted" << std::endl;

    return NO_ERROR;
}

///--------------------------------------------------------------------------------------------------
/// Purpose:
///     @brief Prints usage information
///
/// Returns:
///     @returns none
///--------------------------------------------------------------------------------------------------
void usage(char *name)
{
    std::cout << "Usage:\t" << name << " -U username -D domain -P password" << std::endl;
    std::cout << "or\t" << name << " -P pin" << std::endl;
    std::cout << "or\t" << name << " -L" << std::endl;
    std::cout << "    -U,  --username <username>                  username" << std::endl;
    std::cout << "    -P,  --password <password>                  password or PIN" << std::endl;
    std::cout << "    -D,  --domain <domain>                      domain" << std::endl;
    std::cout << "    -L,  --logoff                               logoff" << std::endl;
}

static struct option long_options[] = {
{ "help",           no_argument,        0, '?' },
{ "username",       required_argument,  0, 'U' },
{ "password",       required_argument,  0, 'P' },
{ "domain",         required_argument,  0, 'D' },
{ "logoff",         no_argument,        0, 'L' },
{ "debug",          no_argument,        0, 'd' },
{ 0, 0, 0, 0 }
};

///--------------------------------------------------------------------------------------------------
/// Purpose:
///     @brief Application entry point
///
/// Returns:
///     @returns NO_ERROR on success, an error code otherwise
///--------------------------------------------------------------------------------------------------
int main(int argc, char *argv[])
{
    std::string userName, domain, password;
    int ret = NO_ERROR;

    // Process all the options
    while ( 1 )
    {
        int c, option_index = 0;

        /* Process arguments.  Option values are saved as STL strings
             * before blank_it() destroys the originals.
             */
        c = getopt_long(argc, argv, "h?dLU:P:D:",
                        long_options, &option_index);
        if (debug) std::cout << "c:" << c << std::endl;

        if (c == -1)
        {
            break;
        }

        switch (c)
        {
        case 0:
            break;
        case 'h':
        case '?':
            usage(argv[0]);
            return 0;
        case 'U':
            userName = readCredential();
            break;
        case 'P':
            password = readCredential();
            break;
        case 'D':
            domain   = readCredential();
            break;
        case 'L':
            return CitrixSSOnSDK::LogoffSsoUser();
        case 'd':
            debug = 1;
            break;
        default:
            std::cerr << "Unrecognised option";
            usage(argv[0]);
            return INVALID_ARGUMENTS;
        }
    }

    // Insert credentials to SSOn.
    ret = insertSSOnCredentials(userName, domain, password);

    if (ret == INVALID_ARGUMENTS)
    {
        usage(argv[0]);
    }

    return ret;
}
